// -*- tab-width:2 ; indent-tabs-mode:nil -*-
//:: case QueueApplication

/* To verify this example:

  vct --silver=silicon_qp --check-history HistoryQueue.java SpecifiedThread.java HistoryWorker.java AbstractQueue.java
  
 */
//begin(all)
class Sender extends Thread {

  AbstractQueue queue;
  int input[];
  //@ seq<int> vals;
  
  /*@
    resource preFork(frac p)=p==write
      ** Value(this.queue) ** queue!=null ** Value(queue.hist)
      ** Value(this.input) ** PointsTo(queue.hist_active,1/4,true)
      ** Value(this.vals)
      ** input.length==|vals|
      ** (\forall* int i; 0 <= i && i < |vals| ; Perm(input[i],1))
      ** (\forall  int i; 0 <= i && i < |vals| ; input[i]==vals[i])
      ** Hist(queue.hist,1/2,empty)
      ** true;
      
    resource postJoin(frac p)=p==write
      ** Value(this.queue) ** queue!=null ** Value(queue.hist)
      ** Value(this.input) ** PointsTo(queue.hist_active,1/4,true)
      ** Value(this.vals)
      ** input.length==|vals|
      ** (\forall* int i; 0 <= i && i < |vals| ; Perm(input[i],1))
      ** (\forall  int i; 0 <= i && i < |vals| ; input[i]==vals[i])
      ** Hist(queue.hist,1/2,queue.hist.put_all(vals))
      ** true;
  @*/
  
  /*@
    given    seq<int>vals;
    requires input.length==|vals| ** queue!=null
      ** Value(queue.hist) ** Hist(queue.hist,1/2,empty)
      ** PointsTo(queue.hist_active,1/4,true)
      ** (\forall* int i; 0 <= i && i < |vals| ; Perm(input[i],1))
      ** (\forall  int i; 0 <= i && i < |vals| ; input[i]==vals[i]);
    ensures  Value(this.queue) ** queue!=null
      ** this.queue==queue
      ** Value(this.input) ** this.input==input
      ** Value(this.vals) ** this.vals==vals ** preFork(1);
  @*/
  public Sender(AbstractQueue queue,int[] input){
    this.queue=queue;
    this.input=input;
    //@ this.vals=vals;
    //@ fold this.preFork@Thread(1);
    //@ fold this.preFork@Sender(1);
  }
  
  public void run(){
    //@ unfold preFork@Sender(1);

    int N=input.length;
    int i=0;
    
    //@ assert Hist(queue.hist,1/2,queue.hist.put_all(take(vals,i)));
    
    
    /*@
      loop_invariant 0 <= i ** i <= N 
      ** Value(this.queue) ** queue!=null ** Value(queue.hist)
      ** Value(this.input) ** PointsTo(queue.hist_active,1/4,true)
      ** Value(this.vals)
      ** N==input.length ** N==|vals|
      ** (\forall* int k; 0 <= k < |vals| ; Perm(input[k],1))
      ** (\forall  int k; 0 <= k < |vals| ; input[k]==vals[k])
      ** Hist(queue.hist,1/2,queue.hist.put_all(take(vals,i)));
    @*/
    while(i<N){
      queue.put(input[i])
      /*@ with { p=1/2 ; P = queue.hist.put_all(take(vals,i));} @*/;
      //@ assert Hist(queue.hist,1/2,
      //@    queue.hist.put_all(take(vals,i)+seq<int>{vals[i]}));
      //@ assert take(vals,i)+seq<int>{vals[i]}==take(vals,i+1);
      i=i+1;
      //@ assert Hist(queue.hist,1/2,queue.hist.put_all(take(vals,i)));
    }
    //@ assert Hist(queue.hist,1/2,queue.hist.put_all(take(vals,N)));
    //@ assert take(vals,N)==vals;
    //@ fold this.postJoin@Thread(1);
    //@ fold this.postJoin@Sender(1);

  }
  

}


class Receiver extends Thread {

  AbstractQueue queue;
  int output[];
  //@ seq<int> vals;
  
  /*@
    resource preFork(frac p)=p==write
      ** Value(this.queue) ** queue!=null ** Value(queue.hist)
      ** Value(this.output) ** PointsTo(queue.hist_active,1/4,true)
      ** Perm(this.vals,1)
      ** (\forall* int i; 0 <= i && i < output.length ; Perm(output[i],1))
      ** Hist(queue.hist,1/2,empty)
      ** true;
      
    resource postJoin(frac p)=p==write
      ** Value(this.queue) ** queue!=null ** Value(queue.hist)
      ** Value(this.output) ** PointsTo(queue.hist_active,1/4,true)
      ** Perm(this.vals,1)
      ** output.length==|vals|
      ** (\forall* int i; 0 <= i < |vals| ; Perm(output[i],1))
      ** (\forall  int i; 0 <= i < |vals| ; output[i]==vals[i])
      ** Hist(queue.hist,1/2,queue.hist.get_all(vals))
      ** true;
  @*/
  
  /*@
    requires queue!=null ** Value(queue.hist)
      ** Hist(queue.hist,1/2,empty)
      ** PointsTo(queue.hist_active,1/4,true)
      ** (\forall* int i; 0 <= i < output.length ; Perm(output[i],1));
    ensures  Value(this.queue) ** queue!=null
      ** this.queue==queue ** Value(this.output)
      ** this.output==output ** preFork(1);
  @*/
  public Receiver(AbstractQueue queue,int[] output){
    this.queue=queue;
    this.output=output;
    //@ fold this.preFork@Thread(1);
    //@ fold this.preFork@Receiver(1);
  }

// begin(receiver)  
  public void run(){
    //@ unfold preFork@Receiver(1); // skip(receiver)
    int N=output.length, i=0;  //@ vals=seq<int>{};
    /*@ loop_invariant Value(queue) ** Value(queue.hist)
    ** Value(output) ** Perm(vals,1) ** 0 <= i <= N ** i==|vals|
    ** PointsTo(queue.hist_active,1/4,true) ** N==output.length
    ** (\forall* int k; 0 <= k < N ; Perm(output[k],1))
    ** (\forall int k; 0 <= k < i ; output[k]==vals[k])
    ** Hist(queue.hist,1/2,queue.hist.get_all(vals)); @*/
    while(i<N){
      output[i]=queue.get()/*@ with {
          p=1/2 ; P = queue.hist.get_all(vals);} @*/;
      //@ vals=vals+seq<int>{output[i]};
      i=i+1;
    }
    //@ fold this.postJoin@Thread(1); // skip(receiver)
    //@ fold this.postJoin@Receiver(1); // skip(receiver)
  }
// end(receiver)
}
//end(all)

