// -*- tab-width:2 ; indent-tabs-mode:nil -*-
//:: case IncrementHistory
//:: tool silicon
//:: option --check-history

/* To verify this example:

  vct --silver=silicon --check-history IncrementHist.java

 */

class History {

  /*@
    resource inv()=HPerm(x,1);
  @*/

  int x;
  
  /*@
  modifies x;
  ensures  x==\old(x)+1;
  process incr();
  
  modifies x;
  requires n>=0;
  ensures  x==\old(x)+n;
  process  single(int n)=n>0?incr()*single(n-1):empty;
  
  modifies x;
  requires n>=0 && m >= 0;
  ensures  x==\old(x)+n+m;
  process  dual(int n,int m)=single(n)||single(m);

  requires n >= 0 ;
  ensures  single(n)==(n>0?single(n-1)*incr():empty);
  void lemma(int n){
    if (n>=2) {
        lemma(n-2);
    }
  }
  
  requires m>=0 && n >= 0 ;
  ensures single(m)*single(n)==single(m+n);
  void single_axiom(int m,int n);
  @*/
  
}

class SubjectLock {

  /*@
    ensures Value(this.subject) ** this.subject==s;
  @*/
  public SubjectLock(History s){
    subject=s;
  }

  History subject;
  /*@
    resource inv()=Value(subject) ** subject.inv();
  @*/
  
  /*@
    resource valid();
    resource locked(frac p,int count);
  @*/
  
  /*@
    requires inv();
    ensures  valid();
  @*/
  public void commit();

  /*@
    requires valid();
    ensures  inv();
  @*/
  public void uncommit();
  
  
  /*@
    given int count;
    given frac p;
    requires count>=0;
    requires count>0?locked(p,count):[p]valid();
    ensures  locked(p,count+1);
    ensures  count==0 ==> inv();
  @*/
  public void lock();
  
  /*@
    given int count;
    given frac p;
    requires count>0;
    requires locked(p,count);
    requires count==1 ==> inv();
    ensures  count==1?[p]valid():locked(p,count-1);
  @*/
  public void unlock();

}

public class Thread {

  //@ public resource joinToken(frac p);

  //@ public resource preFork(frac p)=true;

  //@ public resource postJoin(frac p)=true;
  
  public Thread(){
    //@ assume false;
  }
  
  /*@
    requires preFork(1);
    ensures  postJoin(1);
  @*/
  public void run();

  /*@
    requires preFork(1);
    ensures  joinToken(1);
  @*/
  public final void start();
  
  /*@
    given frac p;
    requires joinToken(p);
    ensures  postJoin(p);
  @*/
  public final void join();

}


class Worker extends Thread {

  SubjectLock l;
  History s;
  
  /*@
    resource preFork(frac p)=p==write ** Value(this.l)**Value(this.l.subject)**Value(this.s)
      **this.l.subject==this.s**([1/2]this.l.valid())**Hist(s,1/2,empty);
      
    resource postJoin(frac p)=p==write ** Value(this.l)**Value(this.l.subject)**Value(this.s)
      **this.l.subject==this.s**([1/2]this.l.valid())**Hist(s,1/2,s.single(1));
  @*/
  
  /*@
    requires Value(l.subject)**l.subject==s**([1/2]l.valid())**Hist(s,1/2,empty);
    ensures  Value(this.l)**this.l==l;
    ensures  Value(this.s)**this.s==s;
    ensures  preFork(1);
  @*/
  public Worker(History s,SubjectLock l){
    this.l=l;
    this.s=s;
    //@ fold this.preFork@Thread(1);
    //@ fold this.preFork@Worker(1);
  }
  
  public void run(){
    //@ unfold preFork@Worker(1);

    l.lock()/*@ with { p=1/2; count=0; }@*/;
    //@ unfold l.inv();
    //@ unfold l.subject.inv();
    
    //@ assert Hist(s,1/2,empty);
    {
      //@ action s , 1/2 , empty , s.incr();
      s.x=s.x+1;
    }
    //@ assert Hist(s,1/2,s.single(1));

    //@ fold l.subject.inv();
    //@ fold l.inv();
    l.unlock()/*@ with { p=1/2; count=1; }@*/;
    
    //@ assert Hist(s,1/2,s.single(1));

    //@ fold this.postJoin@Thread(1);
    //@ fold this.postJoin@Worker(1);

  }
  

}

public class Main {

  public static void main(){
    History s=new History();
    s.x=37;
    //@ create s;
    //@ split s, 1/2, empty, 1/2, empty;
    //@ fold s.inv();
    SubjectLock lock=new SubjectLock(s);
    //@ fold lock.inv();
    lock.commit();
    
    Worker w1=new Worker(s,lock);
    Worker w2=new Worker(s,lock);
    w1.start();
    w2.start();
    w1.join()/*@ with { p = 1; } @*/;
    w2.join()/*@ with { p = 1; } @*/;
    //@ open w1.postJoin@Worker(1);
    //@ open w2.postJoin@Worker(1);
    //@ unfold w1.postJoin@Worker(1);
    //@ unfold w2.postJoin@Worker(1);
    
    lock.uncommit();
    //@ unfold lock.inv();
    //@ unfold s.inv();
    //@ merge s, 1/2, s.single(1), 1/2, s.single(1);
    //@ destroy s , s.dual(1,1);
    //@ assert s.x==39;
  }

}

