import java.io.Serializable;
import java.security.InvalidParameterException;
import java.util.concurrent.atomic.AtomicInteger;

public class FactoryImpl extends ActiveObject {
    private final AtomicInteger objectIdCounter;

    public FactoryImpl() {
        super(MPAO.FACTORY_ID);
        this.objectIdCounter = new AtomicInteger();
    }

    private int genObjectId() {
        return this.objectIdCounter.incrementAndGet();
    }

    public MathRemote newMath() {
        int objectId = this.genObjectId();
        MathImpl local = new MathImpl(objectId);
        local.start();
        return new MathRemote(MPAO.getWorkerId(), objectId);
    }

    public MergeSortRemote newMergeSort() {
        int objectId = this.genObjectId();
        MergeSortImpl local = new MergeSortImpl(objectId);
        local.start();
        return new MergeSortRemote(MPAO.getWorkerId(), objectId);
    }

    @Override
    protected Serializable callMethod(Call call) {
        // Serializable[] parameters = call.getParameters();
        switch (call.getSignature()) {
        case "newMath":
            return this.newMath();
        case "newMergeSort":
            return this.newMergeSort();
        default:
            throw new InvalidParameterException();
        }
    }

    @Override
    public FactoryRemote getRemote() {
        return new FactoryRemote(MPAO.getWorkerId());
    }
}
