import java.io.Serializable;

import mpi.Request;

/**
 * Represents the result of an active object method call.
 */
public class MPAOResult {
    private final Request request;
    private final Object[] buffer;
    
    /*@
        public resource getToken() =
            Value(this.request) ** request.waitToken() **
            Value(this.buffer);
    @*/

    /*@
        requires Value(request.buf) ** request.buf == buffer;
        requires request.waitToken();
        ensures Value(request) ** Value(buffer);
        ensures getToken();
    @*/
    public MPAOResult(Request request, Object[] buffer) {
        this.request = request;
        this.buffer = buffer;
    }

    /**
     * Get the result of an active object method call.
     *
     * This method blocks until the result of the active object method call
     * is received.
     *
     * Only one thread is allowed to call this method at the same time.
     *
     * @return the result of the active object method call.
     */
// begin(get) begin(getSpec)
    /*@
        given Future fut;
        given frac p;
        given process P(msg);
        requires fut != null ** p != none;
        requires Value(this.request);
        requires Value(this.request.source) ** Value(this.request.dest) **
            Value(this.request.tag) ** Value(this.request.r) **
            Value(this.request.buf);
        requires getToken();
        requires Future(fut, p,
                \sum(Object msg;; waitrecv(this.request.source, this.request.dest,
                this.request.msg, this.request.tag, this.request.r) * P(msg)));
        ensures Value(this.request.source) ** Value(this.request.dest) **
            Value(this.request.tag);
        ensures \result == msg;
        ensures valid(this.request.source, this.request.dest, \result,
            this.request.tag);
        ensures Future(fut, p, P(msg));
    @*/
    public Serializable get() {
        //end(getSpec)
        request.Wait();

        return (Serializable) buffer[0];
        //begin(getSpec)
    }
// end(get) end(getSpec)
}