import java.util.Random;

import mpi.MPI;

public class MergeSortMain {
    public static void main(String[] args) {
        MPI.Init(args);
        MPAO.init();

        MPI.COMM_WORLD.Barrier();

        Random random = new Random();

        if (MPAO.getWorkerId() == 0) {
            System.out.println("Generating random sequence...");
            int[] array = new int[100000];
            for (int i = 0; i < array.length; i++) {
                array[i] = random.nextInt(10000);
            }
            System.out.println("Generating done.");

            System.out.println("Getting MergeSort object...");
            RemoteFuture<MergeSortRemote> f = MPAO.getFactory(0).newMergeSort();
            System.out.println("Request sent.");
            MergeSortRemote mergeSort = f.get();
            System.out.println("Got MergeSort object.");

            System.out.println("Starting sort...");
            RemoteFuture<int[]> result = mergeSort.sort(array, 1);
            System.out.println("Request sent.");
            int[] sortedArr = result.get();
            boolean sorted = true;
            for (int i = 1; sorted && i < sortedArr.length; i++) {
                sorted = sortedArr[i - 1] <= sortedArr[i];
            }
            System.out.println("Sorted = " + sorted);
        }

        MPI.COMM_WORLD.Barrier();
        MPI.Finalize();
        System.exit(0);
    }
}
